<?php
// CM_TOTP_demo.php - Version 1.0
//
// Sources and inspiration
//https://tools.ietf.org/html/rfc4648
//https://tools.ietf.org/html/rfc6238
//https://github.com/PHPGangsta/GoogleAuthenticator

// Read more here about this version
// https://support.countermail.com/kb/faq.php?id=259

//Example usuage below, this could of course be extended with much more features other bling bling!

$totp = new TOTP();

// Some examples with pages that uses TOTP authentication
$secrets = array("Stripe.com"    => "Enter the secret code you see the first time only, during activation",
                "LocalBitCoins"  => "Enter the secret code you see the first time only, during activation",
                "PayPal"         => "Enter the secret code you see the first time only, during activation",
                );

/*
 Comment QR-images can be converted to secret text code! 
 The mobile app: "QR CODE READER - FREE" used to work
*/

foreach ($secrets as $key => $value)
{
   $code = $totp->generateTOTP(strtoupper($value));
   echo "\nHomepage: $key:\nTOTP Code:\n$code\n";
}


/* ---------------------------------------------------
 How to verify on server side:
*/

//$result = $totp->verifyTOTP($secret, $code, 2);
//echo $result:
/* --------------------------------------------------- */


//Clearing up the RAM a bit, before exiting, better safe than sorry
$code ="101dummycodedummycode101";
$code = $totp->generateTOTP(strtoupper($code));
unset($code);
unset($secrets);
unset($totp);

exit(1);

//
// The TOTP CLASS
//
class TOTP
{
   private $hash_algo = 'sha1';
   private $digits = 6;

   public function __destruct()
   {
      //$this->challenge_code = null;
   }

   public function __construct()
   {
      //$this->challenge_code = null;
   }

   public function generateSecret($length = 16)
   {
      $bytes = openssl_random_pseudo_bytes($length, $cstrong);

      //Pseudosecure enough?
      if( $cstrong !== true)
      {
         throw new Exception("Failed to generate cryptography strong secret");
      }

      $validChars = $this->getBase32LookupTable();

      $secret = '';
      for ($a = 0; $a < $length; $a++)
      {
         $secret .= $validChars[ord($bytes[$a]) & 31];
      }

      return $secret;
   }
   //FIXME: $key could it be a class variable? $this->key

   /**
    * Returns the code as a string
    */
   public function generateTOTP($secret, $timeSlice = null)
   {
      if ($timeSlice === null)
      {
         $timeSlice = floor(time() / 30);
      }

      $secret = $this->base32Decode($secret);

      $time = chr(0).chr(0).chr(0).chr(0).pack('N*', $timeSlice);

      $hhmac = hash_hmac($this->hash_algo, $time, $secret, true);

      $offset = ord(substr($hhmac, -1)) & 0x0F;

      $hashpart = substr($hhmac, $offset, 4);

      //unpack binary value
      $value = unpack('N', $hashpart);
      $value = $value[1];

      //vi vill ha 32 bitar (0x7FFFFFFF)
      $value = $value & 0x7FFFFFFF;

      $modulo = pow(10, $this->digits);
      $value = $value % $modulo;

      return str_pad($value, $this->digits, '0', STR_PAD_LEFT);
   }

   public function verifyTOTP($secret, $code, $discrepancy = 1, $timeSlice = null)
   {
      if ($timeSlice === null)
      {
         $timeSlice = floor(time() / 30);
      }

      if (strlen($code) != 6)
      {
         throw new Exception('The code is invalid');
      }

      //loopa igenom lite före och efter ifall tiden diffar på enheterna, default är +- 30 sekunder diff
      for ($a = -$discrepancy; $a <= $discrepancy; $a++)
      {
         $calculatedCode = $this->generateTOTP($secret, $timeSlice + $a);

        // if (hash('sha256', $code) == hash('sha256', $calculatedCode)) //hasha för timing attacks?
         if ($code == $calculatedCode)
         {
            return true;
         }
      }

      return false;
      //throw new Exception('The code is invalid 2');
   }

   private function getBase32LookupTable()
   {
      return array(
         'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', //  7
         'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', // 15
         'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', // 23
         'Y', 'Z', '2', '3', '4', '5', '6', '7', // 31
         '=',  // padding char
      );
   }

   private function base32Decode($secret)
   {
      if (empty($secret))
      {
         return '';
      }

      $base32chars = $this->getBase32LookupTable();
      $base32charsFlipped = array_flip($base32chars);
      $paddingCharCount = substr_count($secret, $base32chars[32]);
      $allowedValues = array(6, 4, 3, 1, 0);

      if (!in_array($paddingCharCount, $allowedValues))
      {
         return false;
      }

      for ($i = 0; $i < 4; ++$i)
      {
         if ($paddingCharCount == $allowedValues[$i] && substr($secret, -($allowedValues[$i])) != str_repeat($base32chars[32], $allowedValues[$i]))
         {
            return false;
         }
      }

      $secret = str_replace('=', '', $secret);
      $secret = str_split($secret);
      $binaryString = '';

      $length_secret = count($secret);
      for ($i = 0; $i < $length_secret ; $i = $i + 8)
      {
         $x = '';
         if (!in_array($secret[$i], $base32chars))
         {
            return false;
         }

         for ($j = 0; $j < 8; ++$j)
         {
            $x .= str_pad(base_convert(@$base32charsFlipped[@$secret[$i + $j]], 10, 2), 5, '0', STR_PAD_LEFT);
         }

         $eightBits = str_split($x, 8);

         $length_eightBits = count($eightBits);
         for ($z = 0; $z < $length_eightBits; ++$z)
         {
            $binaryString .= (($y = chr(base_convert($eightBits[$z], 2, 10))) || ord($y) == 48) ? $y : '';
         }
      }
      return $binaryString;
   }
}
//
?>
